# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import json
import tarfile
from pathlib import Path
import tempfile
import numpy as np
from time import sleep

import xobjects as xo
import xtrack as xt

from .user import get_domain, get_directory
from .server import fs_mv, missing_eos, timestamp
from .simulation_io import XbInput, app_version, assert_versions


_line_options = {
    'store_element_names':          False,
    'remove_markers':               True,
    'remove_zero_length_drifts':    True,
    'remove_inactive_multipoles':   True,
    'remove_redundant_apertures':   True,
    'merge_consecutive_multipoles': True,
    'merge_consecutive_drifts':     True,
    'use_simple_bends':             False,
    'use_simple_quadrupoles':       False
}

def _preprocess_line(line, options):
    if line is None:
        return None
    has_tracker = False
    this_options = options.copy()
    this_options.pop('store_element_names')
    if not any(this_options.values()):
        # No reason to discard and rebuild tracker
        return line
    if line.tracker is not None:
        has_tracker = True
        buffer = line._buffer
        io_buffer = line.tracker.io_buffer
        self.discard_tracker()
    # Optimise
    if not this_options['remove_markers']:
        line.remove_markers()
    if not this_options['remove_zero_length_drifts']:
        line.remove_zero_length_drifts()
    if not this_options['remove_inactive_multipoles']:
        line.remove_inactive_multipoles()
    if this_options['remove_redundant_apertures']:
        line.remove_redundant_apertures()
    if this_options['merge_consecutive_multipoles']:
        line.merge_consecutive_multipoles()
    if this_options['merge_consecutive_drifts']:
        line.merge_consecutive_drifts()
    if this_options['use_simple_bends']:
        line.use_simple_bends()
    if this_options['use_simple_quadrupoles']:
        line.use_simple_quadrupoles()
    # Rebuild tracker if it was present
    if has_tracker:
        line.build_tracker(_buffer=buffer,io_buffer=io_buffer)
    return line


def _get_num_elements_from_line(line):
    if line is None:
        return {}
    elements = np.unique([ee.__class__.__name__ for ee in line.elements],
                         return_counts=True)
    return dict(zip(*elements))


class SubmitJobs:

    def __init__(self, user, study_name, line=None, dev_server=False, **kwargs):
        """
        Parameters
        ----------
        user : string
            The user that submits to BOINC. Make sure all permissions are set
            (the user should be member of the CERN xboinc-submitters egroup).
        study_name : string
            The name of the study. This will go inside the job jsons and the
            filenames of the tars.
        line : xtrack.Line, optional
            The line to be tracked. Can be provided globally at the class
            construction, or for each job separately. The latter is much
            slower as it will be preprocessed at each job addition.
        dev_server: bool, optional
            Whether or not to submit to the dev server. Defaults to False.

        Additionally, the following optimisation flags can be passed
        globally:
            store_element_names : default False.
            remove_markers : default True.
            remove_zero_length_drifts : default True.
            remove_inactive_multipoles : default True.
            remove_redundant_apertures : default True.
            merge_consecutive_multipoles : default True.
            merge_consecutive_drifts : default True.
            use_simple_bends : default False.
            use_simple_quadrupoles : default False.

        Usage
        -----
        Create one SubmitJobs instance per study, add jobs one-by-one with
        SubmitJobs.add(), and submit with SubmitJobs.submit().
        """

        assert_versions()
        if not dev_server:
            raise NotImplementedError("Regular server not yet operational. "
                                    + "Please use dev_server=True.")
        if '__' in study_name:
            raise ValueError("The character sequence '__' is not allowed in 'study_name'!")
        self._user = user
        self._domain = get_domain(user)
        if self._domain=='eos':
            missing_eos()
        if dev_server:
            self._target = get_directory(user) / 'input_dev'
        else:
            self._target = get_directory(user) / 'input'
        self._study_name = study_name
        self._line_options = {k: kwargs.pop(k, v) for k, v in _line_options.items()}
        self._line = _preprocess_line(line, self._line_options)
        self._num_elements = _get_num_elements_from_line(line)
        self._submit_file = f"{self._user}__{self._study_name}__{timestamp()}.tar.gz"
        self._json_files = []
        self._bin_files = []
        self._temp    = tempfile.TemporaryDirectory()
        self._tempdir = Path(self._temp.name).resolve()
        self._tempdir.mkdir(parents=True, exist_ok=True)
        self._submitted = False


    def _assert_not_submitted(self):
        if self._submitted:
            raise ValueError("Jobs already submitted! Make a new SubmitJobs object to continue.")


    def add(self, *, job_name, num_turns, particles, line=None, checkpoint_every=-1, **kwargs):
        """
        Add a single job to the SubmitJobs instance. This will create a binary input file and a
        json file (with the same name) containing the job metadata.

        Parameters
        ----------
        job_name : dict
            Name of this individual job.
        num_turns : int
            The number of turns this job should track.
        particles : xpart.Particles
            The particles to be tracked.
        line : xtrack.Line, optional
            The line to be tracked. Can be provided globally at the class
            construction, or here, for each job separately. The latter is
            much slower as it will be preprocessed at each job addition.
        checkpoint_every : int, optional
            When to checkpoint. The default value -1 represents no
            checkpointing.

        Returns
        -------
        None.
        """

        self._assert_not_submitted()
        if '__' in job_name:
            raise ValueError("The character sequence '__' is not allowed in 'job_name'!")

        # Get the line options from kwargs, and default to the options in SubmitJobs
        _line_options = {k: kwargs.pop(k, v) for k, v in self._line_options.items()}

        # Get the line from kwargs, and default to the line in SubmitJobs
        if line is None:
            if self._line is None:
                raise ValueError("Need to provide a line! This can be done for "
                               + "each job separately, or at the SubmitJobs init.")
            if not xt.line._dicts_equal(_line_options, self._line_options):
                raise ValueError(f"Some different line options are given to job "
                               + f"{job_name}, compared to the options in SubmitJobs, "
                               + f"but no new line given!")
            line = self._line
            num_elements = self._num_elements
        else:
            # If a new line is given, preprocess it
            line = _preprocess_line(line, _line_options)
            num_elements = _get_num_elements_from_line(line)

        sleep(0.001) # To enforce different filenames
        filename = f"{self._user}__{timestamp(ms=True)}"
        json_file = Path(self._tempdir, f"{filename}.json")
        bin_file  = Path(self._tempdir, f"{filename}.bin")
        # TODO: warn if job expected to be too short ( < 90s)
        json_dict = {
            'user': self._user,
            'study_name': self._study_name,
            'job_name': job_name,
            'xboinc_ver': app_version,
            'num_elements': num_elements,
            'num_part': len(particles.state[particles.state > 0]),
            'num_turns': num_turns,
            **kwargs
        }
        with json_file.open('w') as fid:
            json.dump(json_dict, fid, cls=xo.JEncoder)
        store_element_names = _line_options['store_element_names']
        data = XbInput(num_turns=num_turns, line=line, checkpoint_every=checkpoint_every,
                       particles=particles, store_element_names=store_element_names)
        data.to_binary(bin_file)
        self._json_files += [json_file]
        self._bin_files  += [bin_file]


    def submit(self):
        """
        Zip all files into a tarfile, and move it to the dedicated user
        folder for submission, which the BOINC server will periodically
        query for new submissions.

        Parameters
        ----------
        None.

        Returns
        -------
        None.
        """

        self._assert_not_submitted()
        with tarfile.open(self._tempdir / self._submit_file, "w:gz") as tar:
            for thisfile in self._json_files + self._bin_files:
                tar.add(thisfile, arcname=thisfile.name)
        if self._domain == 'eos' or self._domain == 'afs':
            fs_mv(self._tempdir / self._submit_file, self._target)
        else:
            raise ValueError(f"Wrong domain {self._domain} for user {self._user}!")
        self._submitted = True
        # TODO: check that tar contains all files
        # clean up
        for thisfile in self._json_files + self._bin_files:
            thisfile.unlink()
        self._temp.cleanup()




# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

# ===============================================================================================
# IMPORTANT
# ===============================================================================================
# Only make changes to this file just before a minor version bump (need a separate commit though)
# to avoid having multiple xboinc versions with out-of-sync executables.
# ===============================================================================================

from pathlib import Path
import numpy as np

import xobjects as xo
import xpart as xp

from .version import XbVersion, assert_versions


class XbState(xo.Struct):
    _version   = XbVersion    # This HAS to be the first field!
    _i_turn    = xo.Int64      # Current turn in tracking (not necessarily the same as particles.at_turn)
    _xsize     = xo.Int64      # Needed to have access to the size in C
    _particles = xp.Particles._XoStruct

    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        particles : xpart.Particles
            The particles in their binary state.
        """

        assert_versions()
        kwargs['_version'] = XbVersion()
        particles = kwargs.pop('particles', None)
        if particles is None or not isinstance(particles, xp.Particles):
            raise ValueError("Need to provide `particles` to XbState.")
        kwargs['_particles'] = particles._xobject
        super().__init__(**kwargs)
        self._xsize = self._size


    @classmethod
    def from_binary(cls, filename, offset=0, raise_version_error=True):
        """
        Create an XbState from a binary file. The file should not
        contain anything else (otherwise the offset will be wrong).
    
        Parameters
        ----------
        filename : pathlib.Path
            The binary containing the simulation state.
    
        Returns
        -------
        XbState
        """

        # Read binary
        filename = Path(filename)
        with filename.open('rb') as fid:
            state_bytes = fid.read()
        buffer_data = xo.ContextCpu().new_buffer(capacity=len(state_bytes))
        buffer_data.buffer[:] = np.frombuffer(state_bytes, dtype=np.int8)
        # Cast to XbVersion to verify versions of xsuite packages
        version_offset = -1
        for field in cls._fields:
            if field.name == '_version':
                version_offset = field.offset
        if version_offset == -1:
            raise ValueError("No xofield `_version` found in XbState!")
        xb_ver = XbVersion._from_buffer(buffer=buffer_data, offset=offset+version_offset)
        if not xb_ver.assert_version(raise_error=raise_version_error, filename=filename):
            return None
        # Retrieve simulation state
        return cls._from_buffer(buffer=buffer_data, offset=offset)

    def to_binary(self, filename):
        """
        Dump the XbState to a binary file.
    
        Parameters
        ----------
        filename : pathlib.Path
            The binary containing the simulation state.
    
        Returns
        -------
        None.
        """
        assert self._offset == 0     # TODO: create new buffer if this is not the case (like when XbState inside XbInput)
        filename = Path(filename).expanduser().resolve()
        with filename.open('wb') as fid:
            fid.write(self._buffer.buffer.tobytes())


    @property
    def version(self):
        return self._version

    @property
    def particles(self):
        return xp.Particles(_xobject=self._particles)

    @property
    def i_turn(self):
        return self._i_turn




# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from .general import _pkg_root, __version__, __xsuite__versions__

from .register import register, deregister
from .submit import SubmitJobs
from .retrieve import RetrieveJobs

from .simulation_io import XbState, XbInput, app_version, app_version_int, assert_versions
from .executable import generate_executable_source, generate_executable

_skip_xsuite_version_check = False



# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import json
import os
from pathlib import Path

from .general import _pkg_root
from .user import update_user_data, get_user_data, remove_user
from .server import server_account, dropdir, fs_exists, fs_rm, fs_cp,\
                    afs_add_acl, afs_remove_acl, on_afs, on_eos, fs_path, fs_rename


user_data_file = _pkg_root / 'user_data.json'


def _create_json(user, directory, remove=False):
    register = 'deregister' if remove else 'register'
    user_file = Path.cwd() / f'{register}_{user}.json'
    if remove:
        data = {'user': user}
    else:
        directory = fs_path(directory)
        if on_eos(directory):
            domain = 'eos'
        elif on_afs(directory):
            domain = 'afs'
        else:
            raise ValueError(f"Directory {directory} not on EOS nor AFS! "
                            + "The xboinc server will not be able to reach it.")
        data = {'user': user, 'directory': f'{directory}', 'domain': domain}
    with user_file.open('w') as f:
        json.dump(data, f)
    return user_file, data


def _give_rights(directory, domain, acl='rlwikd'):
    if domain == 'eos':
        raise NotImplementedError("Ask CERN IT for explanation...")
    else:
        afs_add_acl(server_account, directory, acl=acl)


def _remove_rights(directory, domain):
    if domain == 'eos':
        raise NotImplementedError("Ask CERN IT for explanation...")
    else:
        afs_remove_acl(server_account, directory)


def register(user, directory):
    """
    Register a user to the BOINC server and dev server, by declaring
    the username and user boinc directory, and giving access rights
    to the BOINC admin process. This is not instantaneous, as the
    BOINC server periodically parses new users.

    Parameters
    ----------
    user : string
        Name of the user to register.
    directory : pathlib.Path
        Dedicated folder that the BOINC server can reach (i.e. on CERN
        AFS or EOS), which will hold the new submissions and results.
        Should not be accessed manually by the user to avoid syncing
        issues.

    Returns
    -------
    None.
    """

    assert_eos_accessible()
    directory = fs_path(directory)
    if not directory.is_dir():
        raise ValueError(f"Directory {directory} not found or not a directory (or no permissions)!")
    user_file, data = _create_json(user, directory)
    # This is a small hack to avoid losing sixtadm ACLs during testing
    acl = 'rlwikda' if user==server_account else 'rlwikd'
    try:
        _give_rights(directory, data['domain'], acl=acl)
    except Exception as e:
        user_file.unlink()
        raise e
    input_dir      = directory / 'input'
    input_dev_dir  = directory / 'input_dev'
    output_dir     = directory / 'output'
    output_dev_dir = directory / 'output_dev'
    input_dir.mkdir(parents=True, exist_ok=True)
    input_dev_dir.mkdir(parents=True, exist_ok=True)
    output_dir.mkdir(parents=True, exist_ok=True)
    output_dev_dir.mkdir(parents=True, exist_ok=True)
    try:
        _give_rights(input_dir, data['domain'], acl=acl)
        _give_rights(input_dev_dir, data['domain'], acl=acl)
        _give_rights(output_dir, data['domain'], acl=acl)
        _give_rights(output_dev_dir, data['domain'], acl=acl)
    except Exception as e:
        user_file.unlink()
        raise e
    if fs_exists(dropdir / f'de{user_file.name}'):
        fs_rm(dropdir / f'de{user_file.name}')
        print("Removed existing deregistration file on server dropdir.")
    if fs_exists(dropdir / f'dev_de{user_file.name}'):
        fs_rm(dropdir / f'dev_de{user_file.name}')
    if fs_exists(dropdir / user_file.name):
        fs_rm(dropdir / user_file.name)
        print("Replaced existing registration file on server dropdir.")
    if fs_exists(dropdir / f'dev_{user_file.name}'):
        fs_rm(dropdir / f'dev_{user_file.name}')
    try:
        fs_cp(user_file, dropdir, maximum_trials=1)
        fs_rename(dropdir / user_file.name, dropdir / f'dev_{user_file.name}', maximum_trials=1)
        fs_cp(user_file, dropdir, maximum_trials=1)
    except:
        user_file.unlink()
        raise Exception(f"Failed to copy register file to server dropdir.\n"
                       + "Do you have permissions?\n"
                       + "If not, add yourself to the CERN xboinc-submitters egroup.")
    user_file.unlink()
    update_user_data(user, data)


def deregister(user):
    """
    Remove a user from the BOINC server and dev server.  This is
    not instantaneous, as the BOINC server periodically parses
    the users to remove.

    Parameters
    ----------
    user : string
        Name of the user to deregister.

    Returns
    -------
    None.
    """

    assert_eos_accessible()
    user_file, _ = _create_json(user, '', remove=True)
    try:
        data = get_user_data(user)
    except ValueError:
        print(f"User {user} not found in user_data.")
    else:
        directory = fs_path(data['directory'])
        # This is a quick check to avoid losing sixtadm ACLs during testing
        if user!=server_account:
            try:
                input_dir      = directory / 'input'
                input_dev_dir  = directory / 'input_dev'
                output_dir     = directory / 'output'
                output_dev_dir = directory / 'output_dev'
                _remove_rights(input_dir, data['domain'])
                _remove_rights(input_dev_dir, data['domain'])
                _remove_rights(output_dir, data['domain'])
                _remove_rights(output_dev_dir, data['domain'])
                _remove_rights(directory, data['domain'])
            except:
                print(f"Warning: could not remove ACL on {data['directory']} for server "
                    + f"account {server_account}!\nPlease do this manually.")
    if fs_exists(dropdir / user_file.name[2:]):
        fs_rm(dropdir / user_file.name[2:])
        print("Removed existing registration file on server dropdir.")
    if fs_exists(dropdir / f'dev_{user_file.name[2:]}'):
        fs_rm(dropdir / f'dev_{user_file.name[2:]}')
    try:
        fs_cp(user_file, dropdir, maximum_trials=1)
        fs_rename(dropdir / user_file.name, dropdir / f'dev_{user_file.name}', maximum_trials=1)
        fs_cp(user_file, dropdir, maximum_trials=1)
    except:
        user_file.unlink()
        raise Exception(f"Failed to copy deregister file to server dropdir.\n"
                      + f"Please inform an xboinc admin to deregister manually.")
    user_file.unlink()
    remove_user(user)



# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import json

from .user import get_directory, get_domain
from .server.tools import untar
from .simulation_io import XbState, assert_versions


class RetrieveJobs:

    def __init__(self, user, study_name, dev_server=False):
        """
        Parameters
        ----------
        user : string
            The user that submitted to BOINC. Make sure all permissions are set
            (the user should be member of the CERN xboinc-submitters egroup).
        study_name : string
            The name of the study to look for.
        dev_server: bool
            Whether or not to retrieve from the dev server. Defaults to False.

        Usage
        -----
        And RetrieveJobs instance works as an interator. Loop over it to
        retrieve a tuple (xpart.Particles, json_file) per finished job.
        """

        if not dev_server:
            raise NotImplementedError("Regular server not yet operational. "
                                    + "Please use dev_server=True.")
        assert_versions()
        self._study_name = study_name
        self._user = user
        self._domain = get_domain(user)
        if self._domain=='eos':
            missing_eos()
        if dev_server:
            self._directory = get_directory(user) / "output_dev"
        else:
            self._directory = get_directory(user) / "output"
        self._to_delete = []
        # Untar all results, also if not from this study, as it is not
        # known yet which study's jobs are in which result tar.
        for tar_file in self._directory.glob('*.tar.gz'):
            untar(tar_file)


    def _clean(self):
        for json_file in self._to_delete:
            json_file.with_suffix('.bin').unlink()
            json_file.unlink()
        # Delete empty folders
        for folder in self._directory.glob('*/'):
            if not any(folder.iterdir()):
                folder.rmdir()


    def __iter__(self):
        # Create an iterator over all results that are from this study only.
        json_files = []
        for json_file in self._directory.glob('*/*.json'):
            with open(json_file, 'r') as json_file_obj:
                json_content = json.load(json_file_obj)
            if json_content.get("study_name") == self._study_name \
            and json_content.get("user") == self._user:
                json_files.append(json_file)
        json_files = set(json_files) - set(self._to_delete)
        if not json_files:
            print(f"Warning: No JSON files found in {self._directory} for "
                + f"{self._user}, study_name {self._study_name}.")
        self._json_files = iter(json_files)
        return self


    def __next__(self):
        # Retrieve the next result
        try:
            json_file = next(self._json_files)
        except StopIteration:
            self._clean()
            raise StopIteration
        else:
            bin_file = json_file.with_suffix('.bin')
            # We parse the result in a try context, which fails if the versions
            # are not compatible. Instead of raising an error, we just continue.
            try:
                with open(json_file, 'r') as json_file_obj:
                    json_content = json.load(json_file_obj)
                result = XbState.from_binary(bin_file, raise_version_error=False)
            except Exception as e:
                print(f"Error loading binary file {bin_file}: {e}")
            else:
                if result is None:
                    # Failed because of incompatible versions
                    # TODO: should we print a warning??? Probably yes..
                    return self.__next__()
                particles = result.particles
                self._to_delete.append(json_file)
                return particles, json_content




# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import os, subprocess
from pathlib import Path
from time import sleep
import random

from .tools import log_debug, log_info, log_error


# Functions to work with the AFS file system
# ==========================================

def on_afs(file):
    parents = list(fs_path(file).parents)
    if len(parents) < 3:
        return False
    else:
        return parents[-3] == Path('/afs/cern.ch')

def afs_add_acl(user, directory, acl='rlwik', cmd=None, is_server=False):
    cmd = 'afs_add_acl' if cmd is None else cmd
    try:
        directory = fs_path(directory)
        cmd = subprocess.run(['fs', 'sa', directory, user, acl],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if cmd.returncode == 0:
            log_info(f"Set ACL '{acl}' on directory {directory} for user {user}.", cmd=cmd, is_server=is_server)
            return 0
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            log_error(f"Failed to set ACL '{acl}' on {directory} for user {user}.\n{stderr}", cmd=cmd, is_server=is_server)
            return 1
    except Exception as e:
        log_error(f"Failed to set ACL '{acl}' on {directory} for user {user}.\n{stderr}", e, cmd=cmd, is_server=is_server)
        return 1


def afs_remove_acl(user, directory, cmd=None, is_server=False):
    cmd = 'afs_remove_acl' if cmd is None else cmd
    try:
        directory = fs_path(directory)
        cmd = subprocess.run(['fs', 'sa', directory, user, 'none'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if cmd.returncode == 0:
            log_info(f"Removed ACL on directory {directory} for user {user}.", cmd=cmd, is_server=is_server)
            return 0
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            log_error(f"Failed to remove ACL on {directory} for user {user}.\n{stderr}", cmd=cmd, is_server=is_server)
            return 1
    except Exception as e:
        log_error(f"Failed to remove ACL on {directory} for user {user}.\n{stderr}", e, cmd=cmd, is_server=is_server)
        return 1


def afs_print_acl(directory, cmd=None, is_server=False):
    cmd = 'afs_print_acl' if cmd is None else cmd
    try:
        directory = fs_path(directory)
        cmd = subprocess.run(['fs', 'la', directory],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if cmd.returncode == 0:
            stdout = cmd.stdout.decode('UTF-8').strip().split('\n')
            log_info(f"{stdout}", cmd=cmd, is_server=is_server)
            return 0
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            log_error(f"Failed to print ACL on {directory} for user {user}.\n{stderr}", cmd=cmd, is_server=is_server)
            return 1
    except Exception as e:
        log_error(f"Failed to print ACL on {directory} for user {user}.\n{stderr}", e, cmd=cmd, is_server=is_server)
        return 1


# Functions to work with the EOS file system
# ==========================================

EOS_MGM_URL = 'root://eosuser.cern.ch'
eos_env = {**os.environ, 'EOS_MGM_URL': EOS_MGM_URL}

def on_eos(file):
    parents = list(fs_path(file).parents)
    if len(parents) < 2:
        return False
    else:
        return parents[-2] == Path('/eos')

def eos_installed():
    try:
        cmd = subprocess.run(['eos', '--version'], stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE, env=eos_env)
        # Temporary hack as the eos command wrongly returns 255
        return cmd.returncode == 0 or cmd.returncode == 255
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False

def missing_eos(message='', cmd=None, is_server=False):
    cmd = 'missing_eos' if cmd is None else cmd
    if eos_installed():
        return 0
    else:
        message = '' if message == '' else f"{message} "
        log_error(f"{message}EOS is not installed on your system.",
                  EnvironmentError(), cmd=cmd, is_server=is_server)
        return 1


def eos_command_for_find():
    # The command `eos find` has the wrong behaviour on new machines (running eos 5.2).
    # For this reason, the command `eos oldfind` has been introduced.
    # However, this command does not exist on the old machines, so in that case we
    # default back to `eos find`.
    cmd = subprocess.run(['eos', 'oldfind'], stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE, env=eos_env)
    if cmd.returncode == 255:
        # Command not found; we are running on a machine running old eos < 5.2
        return 'find'
    else:
        # Command found; we are running on a machine running new eos >= 5.2
        return 'oldfind'

# def xrdcp_installed():
#     try:
#         cmd = subprocess.run(["xrdcp", "--version"], stdout=subprocess.PIPE,
#                              stderr=subprocess.PIPE, check=True)
#         return cmd.returncode == 0
#     except (subprocess.CalledProcessError, FileNotFoundError):
#         return False


# General file IO functions
# =========================

def _dir(directory):
    directory = Path(directory)
    if not directory.is_dir() or f'{directory}'[-1] == '/':
        return f'{directory}'
    else:
        return f'{directory}/'

def fs_path(path):
    path = Path(path).expanduser().resolve()
    return Path(path.as_posix().replace('/eos/home-','/eos/user/'))

def fs_glob(path, pattern, cmd=None, is_server=False):
    cmd = 'fs_glob' if cmd is None else cmd
    path = fs_path(path)
    err_mess = f"Failed fs_glob {pattern} in {path}!"
    try:
        if on_eos(path):
            if missing_eos(err_mess):
                return []
            cmd = subprocess.run(['eos', eos_command_for_find(), '-name', pattern, f'{path}'],
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=eos_env)
        else:
            cmd = subprocess.run(['find', f'{path}', '-name', pattern],
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if cmd.returncode != 0 or cmd.stdout == b'':
            return []
        files = [fs_path(ss) for ss in cmd.stdout.decode('UTF-8').split()]
        return [f for f in files if fs_exists(f)]
    except Exception as e:
        log_error(f"{err_mess}\n", e, cmd=cmd, is_server=is_server)
        return []

def fs_exists(file, cmd=None, is_server=False):
    cmd = 'fs_exists' if cmd is None else cmd
    file = fs_path(file)
    err_mess = f"Failed fs_exists for {file}!"
    try:
        if on_eos(file):
            if missing_eos(err_mess):
                return False
            cmd = subprocess.run(['eos', eos_command_for_find(), '-name', file.name,
                                  f'{file.parent}'], stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE, env=eos_env)
            if cmd.returncode != 0 or cmd.stdout == b'':
                # Command failed, assume file not found
                return False
            files = cmd.stdout.decode('UTF-8').split()
            if len(files) == 0:
                # File not found
                return False
            # If the file is found, we need to ensure that it is not empty
            # This command will list zero-length files
            cmd = subprocess.run(['eos', eos_command_for_find(), '-0', '-name', file.name,
                                  f'{file.parent}'], stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE, env=eos_env)
            if cmd.returncode != 0 or cmd.stdout == b'':
                # Command failed, assume no empty file found
                return True
            return len(cmd.stdout.decode('UTF-8').split()) == 0
        else:
            return file.exists() and file.stat().st_size!=0
    except Exception as e:
        log_error(f"{err_mess}\n", e, cmd=cmd, is_server=is_server)
        return False

def fs_rm(file, cmd=None, is_server=False):
    cmd = 'fs_rm' if cmd is None else cmd
    file = fs_path(file)
    err_mess = f"Failed fs_rm for {file}!"
    try:
        if on_eos(file):
            if missing_eos(err_mess):
                return 1
            cmd = subprocess.run(['eos', 'rm', '-r', _dir(file)],
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                 env=eos_env)
        else:
            cmd = subprocess.run(['rm', '-r', f'{file}'], stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
        if cmd.returncode == 0:
            return 0
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            log_error(f"{err_mess}\n{stderr}", cmd=cmd, is_server=is_server)
            return 1
    except Exception as e:
        log_error(f"{err_mess}\n", e, cmd=cmd, is_server=is_server)
        return 1


# Always use the following functions file by file, not for a list of files

def fs_cp(file, directory, maximum_trials=10, wait=2.7, cmd=None, is_server=False):
    cmd = 'fs_cp' if cmd is None else cmd
    file   = fs_path(file)
    directory = fs_path(directory)
    if file.parent == directory:
        return 0
    if not directory.is_dir():
        log_error(f"Failed fs_cp: {directory} is not a directory!", cmd=cmd, is_server=is_server)
        return 1
    err_mess = f"Failed fs_cp {file} to {directory}!"
    try:
        for i in range(maximum_trials):
            if on_eos(file) or on_eos(directory):
                if missing_eos(err_mess):
                    return 1
                cmd = subprocess.run(['eos', 'cp', f'{file}', _dir(directory)],
                                     stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                     env=eos_env)
            else:
                cmd = subprocess.run(['cp', f'{file}', f'{directory}'],
                                     stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            if cmd.returncode == 0 and fs_exists(directory / file.name):
                return 0
            if i != maximum_trials-1:
                log_debug(f"{err_mess}\nRetrying ({i})..", cmd=cmd, is_server=is_server)
                sleep(abs(wait+random.normalvariate(0, 0.1*wait)))
        log_error(f"{err_mess}\n", cmd=cmd, is_server=is_server)
        if not fs_exists(directory / file.name):
            log_error(f"Command succeeds but destination file is not created!\n",
                      cmd=cmd, is_server=is_server)
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            log_error(f"Command failed: {stderr}\n", cmd=cmd, is_server=is_server)
        log_error("Giving up.", cmd=cmd, is_server=is_server)
        return 1
    except Exception as e:
        log_error(f"{err_mess}\n", e, cmd=cmd, is_server=is_server)
        return 1

def fs_rename(file, new_file, maximum_trials=10, wait=2.7, cmd=None, is_server=False):
    cmd = 'fs_rename' if cmd is None else cmd
    file     = fs_path(file)
    new_file = fs_path(new_file)
    if file.parent != new_file.parent:
        log_error(f"Failed fs_rename: {file} not in the same directory as {new_file}!",
                  cmd=cmd, is_server=is_server)
        return 1
    err_mess = f"Failed fs_rename {file} to {new_file}!"
    try:
        for i in range(maximum_trials):
            if on_eos(file):
                if missing_eos(err_mess):
                    return 1
                cmd = subprocess.run(['eos', 'mv', f'{file}', new_file],
                                     stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                     env=eos_env)
            else:
                cmd = subprocess.run(['mv', f'{file}', new_file],
                                     stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            if cmd.returncode == 0 and fs_exists(new_file):
                return 0
            if i != maximum_trials-1:
                log_debug(f"{err_mess}\nRetrying ({i})..", cmd=cmd, is_server=is_server)
                sleep(abs(wait+random.normalvariate(0, 0.1*wait)))
        log_error(f"{err_mess}\n", cmd=cmd, is_server=is_server)
        if not fs_exists(new_file):
            log_error(f"Command succeeds but destination file is not created!\n",
                      cmd=cmd, is_server=is_server)
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            log_error(f"Command failed: {stderr}\n", cmd=cmd, is_server=is_server)
        log_error("Giving up.", cmd=cmd, is_server=is_server)
        return 1
    except Exception as e:
        log_error(f"{err_mess}\n", e, cmd=cmd, is_server=is_server)
        return 1

def fs_mv(file, directory, maximum_trials=10, wait=2.7, cmd=None, is_server=False):
    cmd = 'fs_mv' if cmd is None else cmd
    new_file = None
    if not directory.is_dir():
        new_file = directory
        directory = directory.parent
    cp_failed = fs_cp(file, directory, maximum_trials, wait, cmd=cmd, is_server=is_server)
    # returncode 0 means success
    if cp_failed: return 1
    if new_file is not None:
        # rename
        mv_failed = fs_rename(directory / file.name, new_file,
                              maximum_trials, wait, cmd=cmd, is_server=is_server)
        if mv_failed: return 1
    if fs_exists(file, cmd=cmd, is_server=is_server):
        return fs_rm(file, cmd=cmd, is_server=is_server)
    return 0





# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from pathlib import Path
import logging as lg
import traceback
import tarfile
import datetime



def timestamp(ms=False, in_filename=True):
    ms = -3 if ms else -7
    if in_filename:
        return datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S.%f")[:ms]
    else:
        return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:ms]


def log_debug(message, cmd=None, is_server=False):
    if is_server:
        cmd = '' if cmd is None else f'  {cmd}:'
        lg.debug(f"DEBUG {timestamp(ms=True, in_filename=False)}"
               + f"{cmd}  {message}")
    else:
        print(message)

def log_info(message, cmd=None, is_server=False):
    if is_server:
        cmd = '' if cmd is None else f'  {cmd}:'
        lg.info(f"INFO  {timestamp(ms=True, in_filename=False)}"
               + f"{cmd}  {message}")
    else:
        print(message)

def log_error(message, e=None, cmd=None, is_server=False):
    if is_server:
        cmd = '' if cmd is None else f'  {cmd}:'
        lg.error(f"ERROR {timestamp(ms=True, in_filename=False)}"
               + f"{cmd}  {message}", exc_info=True)
    else:
        raise Exception(message, e)


def untar(filename, cmd=None, is_server=False):
    cmd = 'untar' if cmd is None else cmd
    try:
        filename = Path(filename)
        thisdir = filename.parent / filename.stem[:-4]
        thisdir.mkdir(parents=True, exist_ok=True)
        with tarfile.open(filename, 'r:gz') as fid:
            fid.extractall(path=thisdir)
        filename.unlink()
        log_debug(f"Extracted {filename}.", cmd=cmd, is_server=is_server)
    except Exception as e:
        log_error(f"Failed extracting {filename}", e, cmd=cmd, is_server=is_server)




# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

server_account = 'sixtadm'

from .tools import timestamp, log_debug, log_info, log_error, untar
from .fs import *
from .paths import *


# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

# ===============================================================================================
# IMPORTANT
# ===============================================================================================
# Only make changes to this file just before a minor version bump (need a separate commit though)
# to avoid having multiple xboinc versions with out-of-sync executables.
# ===============================================================================================

import numpy as np
from pathlib import Path

import xobjects as xo
import xpart as xp
import xtrack as xt

from .version import XbVersion, assert_versions
from .default_tracker import default_element_classes, get_default_config, ElementRefData
from .output import XbState

# TODO: check compilation on office PC
# TODO: line.particle_ref is not dumped nor retrieved... Why is this no issue?
# TODO: update xsuite_versions
# TODO: parity
# TODO: can we cache the view on line?
# TODO: docstrings!!
# TODO: remove xobject

# TODO: Caching does not work as moving elements to buffer does not work correctly
#       Can we cache by making the line_metadata in one buffer which we then always merge to a new one?
#       Input creation should be faster than it is now (~4s)
# The build time of the input file is largely dominated by the rebuilding of the
# ElementRefData. For this reason we cache the line, such that when submitting
# many jobs on the same line only the first job creation takes some time.
_previous_line_cache = {}

_xboinc_context = xo.ContextCpu()


class XbInput(xo.Struct):
    _version         = XbVersion    # This HAS to be the first field!
    num_turns        = xo.Int64
    num_elements     = xo.Int64
    checkpoint_every = xo.Int64
    _parity_check    = xo.Int64      # TODO
    xb_state         = XbState
    line_metadata    = xo.Ref(ElementRefData)

    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        particles : xpart.Particles
            The particles to be tracked.
        xb_state : XbState
            The state of the particles. Use either this parameter or
            particles, not both.
        line : xtrack.Line
            The line to be tracked.
        line_metadata : ElementRefData
            Currently not supported (need to fix bug in xobjects).
        num_turns : Int64
            The number of turns to track
        checkpoint_every : Int64, optional
            When to checkpoint. The default value -1 represents no
            checkpointing.
        store_element_names : bool, optional
            Whether or not to store the element names in the binary.
            Defaults to True.

        The other xofields are generated automatically and will be
        overwritten if provided.
        """

        assert_versions()
        kwargs['_version'] = XbVersion()
        kwargs.setdefault('_buffer', _xboinc_context.new_buffer())
        kwargs.setdefault('checkpoint_every', -1)
        # Pre-build particles / XbState; will be moved to correct buffer at XoStruct init
        particles = kwargs.pop('particles', None)
        xb_state = kwargs.get('xb_state', None)
        if particles is not None:
            if xb_state is not None:
                raise ValueError("Use `xb_state` or `particles`, not both.")
            kwargs['xb_state'] = XbState(particles=particles, _i_turn=0)
        elif xb_state is None or not isinstance(xb_state, XbState):
            raise ValueError("Need to provide `xb_state` or `particles`.")
        # Get the line, build the metadata after building the XoStruct
        # We need to do it like this because the elements are not moved correctly
        line = kwargs.pop('line', None)
        if kwargs.pop('line_metadata', None) is not None:
            raise ValueError("Cannot provide the line metadata directly!")
        store_element_names = kwargs.pop('store_element_names', True)
        super().__init__(**kwargs)
        self.line_metadata = _build_line_metadata(line, _buffer=self._buffer,
                                                  store_element_names=store_element_names)
        self.num_elements = len(line.elements)
        _shrink(self._buffer)


    @classmethod
    def from_binary(cls, filename, offset=0, raise_version_error=True):
        """
        Create an XbInput from a binary file. The file should not
        contain anything else (otherwise the offset will be wrong).
    
        Parameters
        ----------
        filename : pathlib.Path
            The binary containing the simulation state.
    
        Returns
        -------
        XbInput
        """

        # Read binary
        filename = Path(filename)
        with filename.open('rb') as fid:
            state_bytes = fid.read()
        buffer_data = _xboinc_context.new_buffer(capacity=len(state_bytes))
        buffer_data.buffer[:] = np.frombuffer(state_bytes, dtype=np.int8)
        # Cast to XbVersion to verify versions of xsuite packages
        version_offset = -1
        for field in cls._fields:
            if field.name == '_version':
                version_offset = field.offset
        if version_offset == -1:
            raise ValueError("No xofield `_version` found in XbInput!")
        xb_ver = XbVersion._from_buffer(buffer=buffer_data, offset=offset+version_offset)
        if not xb_ver.assert_version(raise_error=raise_version_error, filename=filename):
            return None
        # Retrieve simulation input
        return cls._from_buffer(buffer=buffer_data, offset=offset)

    def to_binary(self, filename):
        """
        Dump the XbInput to a binary file.
    
        Parameters
        ----------
        filename : pathlib.Path
            The binary containing the simulation state.
    
        Returns
        -------
        None.
        """
        _shrink(self._buffer)
        assert self._offset == 0
        filename = Path(filename).expanduser().resolve()
        with filename.open('wb') as fid:
            fid.write(self._buffer.buffer.tobytes())


    @property
    def version(self):
        return self._version

    @property
    def line(self):
        elements = [el._DressingClass(_xobject=el) for el in self.line_metadata.elements]
        names = self.line_metadata.names
        if len(names) == 0:
            n = len(elements)
            digits = int(np.ceil(np.log10(n)))
            names = [f"el_{i:>0{digits}}" for i in range(n)]
        return xt.Line(elements=elements, element_names=names)

    @line.setter
    def line(self, val):
        # Only works as long as line_metadata is an xo.Ref, but we try to avoid this
        raise NotImplementedError

    @property
    def particles(self):
        return self.xb_state.particles


def _build_line_metadata(line, _buffer=None, store_element_names=True):
    # Create the ElementRefData from a given line
    line_id = id(line)
    # TODO: caching currently doesn't work
    _previous_line_cache = {}
    if line_id not in _previous_line_cache:
        _check_config(line)
        _check_compatible_elements(line)
        if _buffer is None:
            _buffer = _xboinc_context.new_buffer()
        names = list(line.element_names) if store_element_names else []
        element_ref_data = ElementRefData(
            elements=len(line.element_names),
            names=names,
            _buffer=_buffer,
        )
        element_ref_data.elements = [
            line.element_dict[name]._xobject for name in line.element_names
        ]
        _previous_line_cache[line_id] = element_ref_data

    return _previous_line_cache[line_id]

def _check_config(line):
    # Check that the present config is on Xboinc
    default_config_hash = get_default_config()
    for key, val in default_config_hash:
        if key not in line.config:
            print(f"Warning: Configuration option `{key}` not found in line.config! "
                + f"Set to Xboinc default `{val}`.")
        elif val != line.config[key]:
            print(f"Warning: Configuration option `{key}` set to `{line.config[key]}` "
                + f"in line.config! Not supported by Xboinc. Overwritten to default `{val}`.")
    for key in set(line.config.keys()) - {k[0] for k in default_config_hash}:
        print(f"Warning: Configuration option `{key}` requested in line.config!"
            + f"Not supported by Xboinc. Ignored.")

def _check_compatible_elements(line):
    # Check that all elements are supported by Xboinc
    default_elements = [d.__name__ for d in default_element_classes]
    for ee in np.unique([ee.__class__.__name__ for ee in line.elements]):
        if ee not in default_elements:
            raise ValueError(f"Element of type {ee} not supported "
                           + f"in this version of xboinc!")


def _shrink(buffer):
    # Shrink a buffer by removing all free capacity
    if buffer.get_free() > 0:
        new_capacity = buffer.capacity - buffer.get_free()
        newbuff = buffer._new_buffer(new_capacity)
        buffer.copy_to_native(
                dest=newbuff, dest_offset=0, source_offset=0, nbytes=new_capacity
            )
        buffer.buffer = newbuff
        buffer.capacity = new_capacity
        buffer.chunks = []



# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from pathlib import Path

afsusr  = Path('/afs/cern.ch/user/s/sixtadm')
eosusr  = Path('/eos/user/s/sixtadm')

eosdir     = eosusr / 'xboinc_server'         # EOS spooldir for server
dropdir    = eosdir / 'drop'                  # only location where users have write access (i.e. to register)

_test_afs = afsusr / 'public' / 'test_xboinc'
_test_eos = eosusr / 'test_xboinc'



# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import xobjects as xo

from ..simulation_io import XbState, XbInput, XbVersion, app_version, get_default_tracker_kernel, assert_versions
from ..general import _pkg_root

from pathlib import Path
import shutil, subprocess, os

# ===============================================================================================
# IMPORTANT
# ===============================================================================================
# Only make changes to this file just before a minor version bump (need a separate commit though)
# to avoid having multiple xboinc versions with out-of-sync executables.
# ===============================================================================================

insert_in_all_files = """
#include <stdio.h>
#ifndef NULL
    #define NULL 0
#endif
"""

_sources = [Path.cwd() / f for f in ["main.c", "Makefile", "xtrack.c", "xtrack.h"]]


def generate_executable_source(*, overwrite=False, _context=None):
    """
    Generate all source files needed to compile the Xboinc executable.

    Parameters
    ----------
    overwrite : bool, optional
        Whether or not to overwrite existing source files.
    
    Returns
    -------
    None
    """

    assert _context is None
    assert_versions()

    if not (Path.cwd() / "xb_input.h").exists() or overwrite:
        # The XbInput source API should not be static, as it has to be exposed to main.
        # TODO: Do we still want to inline this? If yes, we need to adapt xo.specialize_source
        #       to pass the replacement of /*gpufun*/ as an option
        conf = xo.typeutils.default_conf.copy()
        conf['gpufun'] = ''
        xb_input_sources = [
            insert_in_all_files,
            xo.specialize_source(XbVersion._gen_c_api(conf).source,
                                        specialize_for='cpu_serial'),
            xo.specialize_source(XbState._gen_c_api(conf).source,
                                        specialize_for='cpu_serial'),
            xo.specialize_source(XbInput._gen_c_api(conf).source,
                                        specialize_for='cpu_serial'),
        ]
        xb_input_h = '\n'.join(xb_input_sources)
        with (Path.cwd() / "xb_input.h").open('w') as fid:
            fid.write(xb_input_h)

    if not (Path.cwd() / "xtrack_tracker.h").exists() or overwrite:
        track_kernel = get_default_tracker_kernel()
        xtrack_tracker_h = (
            insert_in_all_files + track_kernel.specialized_source)
        with (Path.cwd() / "xtrack_tracker.h").open('w') as fid:
            fid.write(xtrack_tracker_h)

    for file in _sources:
        if not file.exists() or overwrite:
            shutil.copy(_pkg_root / 'executable' / file.name, Path.cwd())


# BOINC executable naming conventions:
# windows_intelx86            Microsoft Windows (98 or later) running on an Intel x86-compatible CPU   (32bit)
# windows_x86_64              Microsoft Windows running on an AMD x86_64 or Intel EM64T CPU
# i686-pc-linux-gnu           Linux running on an Intel x86-compatible CPU    (32bit)
# x86_64-pc-linux-gnu         Linux running on an AMD x86_64 or Intel EM64T CPU
# powerpc-apple-darwin        Mac OS X 10.3 or later running on Motorola PowerPC
# i686-apple-darwin           Mac OS 10.4 or later running on Intel   (32bit)
# x86_64-apple-darwin         Intel 64-bit Mac OS 10.5 or later
# arm64-apple-darwin          Mac OS on M1 or M2
# sparc-sun-solaris2.7        Solaris 2.7 running on a SPARC-compatible CPU
# sparc-sun-solaris           Solaris 2.8 or later running on a SPARC-compatible CPU
# sparc64-sun-solaris         Solaris 2.8 or later running on a SPARC 64-bit CPU
# powerpc64-ps3-linux-gnu     Sony Playstation 3 running Linux
# arm-android-linux-gnu       Android running on ARM
# aarch64-android-linux-gnu   Android running on aarch64
# aarch64-unknown-linux-gnu   Linux running aarch64
# x86_64-pc-freebsd__sse2     Free BSD running on 64 bit X86

def generate_executable(*, keep_source=False, clean=True, boinc_path=None, target=None):
    """
    Generate the Xboinc executable.

    Parameters
    ----------
    keep_source : bool, optional
        Whether or not to keep the source files. Defaults to False.
    clean : bool, optional
        Whether or not to clean the make directory. Defaults to True.
    boinc_path : pathlib.Path, optional
        The path to the local BOINC installation. If none, an executable
        without the BOINC API is generated. Defaults to None.
    target : string, optional
        The target architecture to compile to. Not yet implemented.
    
    Returns
    -------
    None
    """
    assert target is None
    assert_versions()

    # Check boinc path
    if boinc_path is not None:
        boinc_path = Path(boinc_path).expanduser().resolve()
        if not boinc_path.is_dir() or not boinc_path.exists():
            raise RuntimeError(f"BOINC path {boinc_path} does not exist!")
        boinc_api = boinc_path / 'api' / 'libboinc_api.a'
        boinc_lib = boinc_path / 'lib' / 'libboinc.a'
        if not boinc_api.exists():
            raise RuntimeError(f"Cannot find BOINC API {boinc_api}!")
        if not boinc_lib.exists():
            raise RuntimeError(f"Cannot find BOINC LIB {boinc_lib}!")

    config   = Path.cwd() / "xb_input.h"
    tracker  = Path.cwd() / "xtrack_tracker.h"
    if not config.exists() or not tracker.exists() or not all([s.exists() for s in _sources]):
        generate_executable_source()

    # Verify dependencies
    if shutil.which("gcc") is None and shutil.which("clang") is None:
        raise RuntimeError("Neither `gcc` or `clang` are found. Install a C compiler.")
    if shutil.which("g++") is None and shutil.which("clang++") is None:
        raise RuntimeError("Neither `g++` or `clang++` are found. Install a C++ compiler.")
    if shutil.which("make") is None:
        raise RuntimeError("Please install `make` before generating the executable.")
    if boinc_path is not None:
        for dep in ["automake", "m4", "libtool", "autoconf", "pkg-config"]:
            if shutil.which(dep) is None:
                raise RuntimeError(f"Please install `{dep}` before generating the executable.")
    _check_libstd()

    # Create executable name
    app_tag = f'{app_version}'
    cmd = subprocess.run(['uname', '-m'], stdout=subprocess.PIPE)
    if cmd.returncode == 0:
        machine = cmd.stdout.decode('UTF-8').strip().lower()
    else:
        machine = 'none'
    cmd = subprocess.run(['uname', '-s'], stdout=subprocess.PIPE)
    if cmd.returncode == 0:
        thisos = cmd.stdout.decode('UTF-8').strip().lower()
    else:
        thisos = 'none'
    vendor = 'apple' if thisos=='darwin' else 'pc'
    thisos = f'{thisos}-gnu' if thisos=='linux' else thisos
    app_tag += f"-{machine}-{vendor}-{thisos}"

    # Compile!
    if  target is None:
        cmd = subprocess.run(['make', 'clean'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if cmd.returncode != 0:
            print(cmd.stdout.decode('UTF-8').strip())
            stderr = cmd.stderr.decode('UTF-8').strip()
            raise RuntimeError(f"Make clean failed. Stderr:\n {stderr}")
        if boinc_path is None:
            app = 'xboinc_test'
            cmd = subprocess.run(['make', app],
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        else:
            app = 'xboinc'
            cmd = subprocess.run(['make', app], env={**os.environ, 'BOINC_DIR': \
                                                     boinc_path.as_posix()},
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if cmd.returncode != 0:
            print(cmd.stdout.decode('UTF-8').strip())
            stderr = cmd.stderr.decode('UTF-8').strip()
            raise RuntimeError(f"Compilation failed. Stderr:\n {stderr}")
        else:
            print(cmd.stdout.decode('UTF-8').strip())
            Path(app).rename(f"{app}_{app_tag}")
    else:
        raise NotImplementedError

    # Clean up
    if clean:
        cmd = subprocess.run(['make', 'clean'],
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if cmd.returncode != 0:
            stderr = cmd.stderr.decode('UTF-8').strip()
            raise RuntimeError(f"Could not run `make clean`. Stderr:\n {stderr}")
    if not keep_source:
        config.unlink()
        tracker.unlink()
        for s in _sources: s.unlink()

    
    # if windows64:
    #     if shutil.which("x86_64-w64-mingw32-gcc") is not None:
    #         compiler = "x86_64-w64-mingw32-gcc"
    #     else:
    #         raise ValueError("Mingw32 not found!")
    #     tag += '-windows_x86_64.exe'
    # elif windows32:
    #     if shutil.which("i686-w64-mingw32-gcc") is not None:
    #         compiler = "i686-w64-mingw32-gcc"
    #     else:
    #         raise ValueError("Mingw32 not found!")
    #     tag += '-windows_intelx86.exe'
    # else:
    #     if shutil.which("gcc") is not None:
    #         compiler = "gcc"
    #     elif shutil.which("clang") is not None:
    #         compiler = "clang"
    #     else:
    #         raise RuntimeError("Neither clang or gcc are found. Install a C compiler.")


def _check_libstd():
    missing_lib = False
    stderr = None

    cmd = subprocess.run(['make', 'libstdc++.a'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if cmd.returncode != 0:
        stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
        missing_lib = True
    lib = Path('libstdc++.a')
    if lib.is_symlink():
        try:
            if not lib.exists() or lib.readlink() == lib:
                missing_lib = True
        except PermissionError:
            missing_lib = True
    elif lib.exists():
        raise ValueError("Something is wrong; `libstdc++.a` should be a symlink but is a regular file...")
    else:
        missing_lib = True

    if missing_lib == True:
        stderr = '' if stderr is None else f"\nStderr:\n {stderr}"
        if lib.exists(): lib.unlink()
        raise RuntimeError(f"Make cannot find `libstdc++.a`. Please install it (e.g. `conda install libstdcxx-ng`).{stderr}")



